@@ -25,6 +25,7 @@
  *	Adam Jackson (ajax@redhat.com)
  *	Peter Hutterer (peter.hutterer@redhat.com)
  *	Oliver McFadden (oliver.mcfadden@nokia.com)
+ *	Thomas H.P. Andersen (phomes@gmail.com)
  */
 
 #ifdef HAVE_CONFIG_H
@@ -53,9 +54,7 @@
 #include <X11/Xatom.h>
 #include <evdev-properties.h>
 #include <xserver-properties.h>
-#ifdef MULTITOUCH
 #include <mtdev-plumbing.h>
-#endif
 
 #ifndef XI_PROP_PRODUCT_ID
 #define XI_PROP_PRODUCT_ID "Device Product ID"
@@ -90,13 +89,6 @@
 #define XI86_SERVER_FD 0x20
 #endif
 
-static const char *evdevDefaults[] = {
-    "XkbRules",     "evdev",
-    "XkbModel",     "pc104", /* the right model for 'us' */
-    "XkbLayout",    "us",
-    NULL
-};
-
 /* Any of those triggers a proximity event */
 static int proximity_bits[] = {
         BTN_TOOL_PEN,
@@ -139,6 +131,7 @@ static int EvdevSwitchMode(ClientPtr client, DeviceIntPtr device, int mode)
 {
     InputInfoPtr pInfo;
     EvdevPtr pEvdev;
+    int val;
 
     pInfo = device->public.devicePrivate;
     pEvdev = pInfo->private;
@@ -154,10 +147,21 @@ static int EvdevSwitchMode(ClientPtr client, DeviceIntPtr device, int mode)
     switch (mode) {
         case Absolute:
             pEvdev->flags &= ~EVDEV_RELATIVE_MODE;
+            if (valuator_mask_fetch(pEvdev->old_vals, 0, &val))
+                valuator_mask_set(pEvdev->abs_vals, 0, val);
+            if (valuator_mask_fetch(pEvdev->old_vals, 1, &val))
+                valuator_mask_set(pEvdev->abs_vals, 1, val);
+            valuator_mask_zero(pEvdev->old_vals);
             break;
 
         case Relative:
             pEvdev->flags |= EVDEV_RELATIVE_MODE;
+            if (valuator_mask_fetch(pEvdev->abs_vals, 0, &val))
+                valuator_mask_set(pEvdev->old_vals, 0, val);
+            if (valuator_mask_fetch(pEvdev->abs_vals, 1, &val))
+                valuator_mask_set(pEvdev->old_vals, 1, val);
+            valuator_mask_unset(pEvdev->abs_vals, 0);
+            valuator_mask_unset(pEvdev->abs_vals, 1);
             break;
 
         default:
@@ -254,12 +258,6 @@ out:
     return rc;
 }
 
-#ifndef HAVE_SMOOTH_SCROLLING
-static int wheel_up_button = 4;
-static int wheel_down_button = 5;
-static int wheel_left_button = 6;
-static int wheel_right_button = 7;
-#endif
 
 static EventQueuePtr
 EvdevNextInQueue(InputInfoPtr pInfo)
@@ -320,7 +318,6 @@ EvdevQueueProximityEvent(InputInfoPtr pInfo, int value)
     }
 }
 
-#ifdef MULTITOUCH
 void
 EvdevQueueTouchEvent(InputInfoPtr pInfo, unsigned int touch, ValuatorMask *mask,
                      uint16_t evtype)
@@ -334,7 +331,6 @@ EvdevQueueTouchEvent(InputInfoPtr pInfo, unsigned int touch, ValuatorMask *mask,
         pQueue->val = evtype;
     }
 }
-#endif
 
 /**
  * Post button event right here, right now.
@@ -434,72 +430,44 @@ static void
 EvdevProcessValuators(InputInfoPtr pInfo)
 {
     EvdevPtr pEvdev = pInfo->private;
-    int *delta = pEvdev->delta;
-
-    if (pEvdev->abs_queued) {
-        /* convert to relative motion for touchpads */
-        if (pEvdev->flags & EVDEV_RELATIVE_MODE) {
-            if (pEvdev->in_proximity) {
-                if (valuator_mask_isset(pEvdev->vals, 0))
-                {
-                    if (valuator_mask_isset(pEvdev->old_vals, 0))
-                        delta[REL_X] = valuator_mask_get(pEvdev->vals, 0) -
-                            valuator_mask_get(pEvdev->old_vals, 0);
-                    valuator_mask_set(pEvdev->old_vals, 0,
-                            valuator_mask_get(pEvdev->vals, 0));
-                }
-                if (valuator_mask_isset(pEvdev->vals, 1))
-                {
-                    if (valuator_mask_isset(pEvdev->old_vals, 1))
-                        delta[REL_Y] = valuator_mask_get(pEvdev->vals, 1) -
-                            valuator_mask_get(pEvdev->old_vals, 1);
-                    valuator_mask_set(pEvdev->old_vals, 1,
-                            valuator_mask_get(pEvdev->vals, 1));
-                }
-            } else {
-                valuator_mask_zero(pEvdev->old_vals);
-            }
-            valuator_mask_zero(pEvdev->vals);
-            pEvdev->abs_queued = 0;
-            pEvdev->rel_queued = 1;
-        } else {
-            int val;
-            if (valuator_mask_fetch(pEvdev->vals, 0, &val))
-                valuator_mask_set(pEvdev->old_vals, 0, val);
-            if (valuator_mask_fetch(pEvdev->vals, 1, &val))
-                valuator_mask_set(pEvdev->old_vals, 1, val);
-        }
-    }
 
+    /* Apply transformations on relative coordinates */
     if (pEvdev->rel_queued) {
-        int tmp;
-        int i;
+        double deltaX = 0, deltaY = 0;
+
+        if (valuator_mask_isset(pEvdev->rel_vals, REL_X))
+            deltaX = valuator_mask_get_double(pEvdev->rel_vals, REL_X);
+        if (valuator_mask_isset(pEvdev->rel_vals, REL_Y))
+            deltaY = valuator_mask_get_double(pEvdev->rel_vals, REL_Y);
 
         if (pEvdev->swap_axes) {
-            tmp = pEvdev->delta[REL_X];
-            pEvdev->delta[REL_X] = pEvdev->delta[REL_Y];
-            pEvdev->delta[REL_Y] = tmp;
-            if (pEvdev->delta[REL_X] == 0)
-                valuator_mask_unset(pEvdev->vals, REL_X);
-            if (pEvdev->delta[REL_Y] == 0)
-                valuator_mask_unset(pEvdev->vals, REL_Y);
+            double tmp = deltaX;
+            deltaX = deltaY;
+            deltaY = tmp;
         }
+
+        if (pEvdev->resolution > 0) {
+            deltaX *= DEFAULT_MOUSE_DPI / pEvdev->resolution;
+            deltaY *= DEFAULT_MOUSE_DPI / pEvdev->resolution;
+        }
+
         if (pEvdev->invert_x)
-            pEvdev->delta[REL_X] *= -1;
+            deltaX *= -1;
         if (pEvdev->invert_y)
-            pEvdev->delta[REL_Y] *= -1;
+            deltaY *= -1;
 
+        if (deltaX)
+            valuator_mask_set_double(pEvdev->rel_vals, REL_X, deltaX);
+        else
+            valuator_mask_unset(pEvdev->rel_vals, REL_X);
 
-        Evdev3BEmuProcessRelMotion(pInfo,
-                                   pEvdev->delta[REL_X],
-                                   pEvdev->delta[REL_Y]);
+        if (deltaY)
+            valuator_mask_set_double(pEvdev->rel_vals, REL_Y, deltaY);
+        else
+            valuator_mask_unset(pEvdev->rel_vals, REL_Y);
+
+        Evdev3BEmuProcessRelMotion(pInfo, deltaX, deltaY);
 
-        for (i = 0; i < REL_CNT; i++)
-        {
-            int map = pEvdev->rel_axis_map[i];
-            if (pEvdev->delta[i] && map != -1)
-                valuator_mask_set(pEvdev->vals, map, pEvdev->delta[i]);
-        }
     }
     /*
      * Some devices only generate valid abs coords when BTN_TOOL_PEN is
@@ -511,9 +479,9 @@ EvdevProcessValuators(InputInfoPtr pInfo)
      * just works.
      */
     else if (pEvdev->abs_queued && pEvdev->in_proximity) {
-        EvdevSwapAbsValuators(pEvdev, pEvdev->vals);
-        EvdevApplyCalibration(pEvdev, pEvdev->vals);
-        Evdev3BEmuProcessAbsMotion(pInfo, pEvdev->vals);
+        EvdevSwapAbsValuators(pEvdev, pEvdev->abs_vals);
+        EvdevApplyCalibration(pEvdev, pEvdev->abs_vals);
+        Evdev3BEmuProcessAbsMotion(pInfo, pEvdev->abs_vals);
     }
 }
 
@@ -559,10 +527,10 @@ EvdevProcessProximityState(InputInfoPtr pInfo)
     if (!pEvdev->prox_queued)
     {
         if (pEvdev->abs_queued && !pEvdev->in_proximity)
-            for (i = 0; i < valuator_mask_size(pEvdev->vals); i++)
-                if (valuator_mask_isset(pEvdev->vals, i))
+            for (i = 0; i < valuator_mask_size(pEvdev->abs_vals); i++)
+                if (valuator_mask_isset(pEvdev->abs_vals, i))
                     valuator_mask_set(pEvdev->prox, i,
-                                      valuator_mask_get(pEvdev->vals, i));
+                                      valuator_mask_get(pEvdev->abs_vals, i));
         return 0;
     }
 
@@ -581,13 +549,13 @@ EvdevProcessProximityState(InputInfoPtr pInfo)
         /* We're about to go into/out of proximity but have no abs events
          * within the EV_SYN. Use the last coordinates we have. */
         for (i = 0; i < valuator_mask_size(pEvdev->prox); i++)
-            if (!valuator_mask_isset(pEvdev->vals, i) &&
+            if (!valuator_mask_isset(pEvdev->abs_vals, i) &&
                 valuator_mask_isset(pEvdev->prox, i))
-                valuator_mask_set(pEvdev->vals, i,
+                valuator_mask_set(pEvdev->abs_vals, i,
                                   valuator_mask_get(pEvdev->prox, i));
         valuator_mask_zero(pEvdev->prox);
 
-        pEvdev->abs_queued = valuator_mask_size(pEvdev->vals);
+        pEvdev->abs_queued = valuator_mask_size(pEvdev->abs_vals);
     }
 
     pEvdev->in_proximity = prox_state;
@@ -639,23 +607,6 @@ EvdevProcessRelativeMotionEvent(InputInfoPtr pInfo, struct input_event *ev)
     value = ev->value;
 
     switch (ev->code) {
-#ifndef HAVE_SMOOTH_SCROLLING
-        case REL_WHEEL:
-            if (value > 0)
-                EvdevQueueButtonClicks(pInfo, wheel_up_button, value);
-            else if (value < 0)
-                EvdevQueueButtonClicks(pInfo, wheel_down_button, -value);
-            break;
-
-        case REL_DIAL:
-        case REL_HWHEEL:
-            if (value > 0)
-                EvdevQueueButtonClicks(pInfo, wheel_right_button, value);
-            else if (value < 0)
-                EvdevQueueButtonClicks(pInfo, wheel_left_button, -value);
-            break;
-        /* We don't post wheel events as axis motion. */
-#endif
         default:
             /* Ignore EV_REL events if we never set up for them. */
             if (!(pEvdev->flags & EVDEV_RELATIVE_EVENTS) &&
@@ -668,14 +619,16 @@ EvdevProcessRelativeMotionEvent(InputInfoPtr pInfo, struct input_event *ev)
                 return;
 
             pEvdev->rel_queued = 1;
-            pEvdev->delta[ev->code] += value;
             map = pEvdev->rel_axis_map[ev->code];
-            valuator_mask_set(pEvdev->vals, map, value);
+
+            if (valuator_mask_isset(pEvdev->rel_vals, map))
+                value += valuator_mask_get(pEvdev->rel_vals, map);
+
+            valuator_mask_set(pEvdev->rel_vals, map, value);
             break;
     }
 }
 
-#ifdef MULTITOUCH
 static void
 EvdevProcessTouch(InputInfoPtr pInfo)
 {
@@ -749,8 +702,19 @@ EvdevProcessTouchEvent(InputInfoPtr pInfo, struct input_event *ev)
         !libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_MT_SLOT))
         return;
 
+    if (pEvdev->fake_mt)
+        return;
+
     if (ev->code == ABS_MT_SLOT) {
         EvdevProcessTouch(pInfo);
+        if (ev->value >= num_slots(pEvdev) ) {
+            LogMessageVerbSigSafe(X_WARNING, 0,
+                                  "%s: Slot index %d out of bounds (max %d), touch events may be incorrect.\n",
+                                  pInfo->name,
+                                  ev->value,
+                                  num_slots(pEvdev) - 1);
+            return;
+        }
         pEvdev->cur_slot = ev->value;
     } else
     {
@@ -780,10 +744,6 @@ EvdevProcessTouchEvent(InputInfoPtr pInfo, struct input_event *ev)
         }
     }
 }
-#else
-#define EvdevProcessTouch(pInfo)
-#define EvdevProcessTouchEvent(pInfo, ev)
-#endif /* MULTITOUCH */
 
 /**
  * Take the absolute motion input event and process it accordingly.
@@ -813,8 +773,20 @@ EvdevProcessAbsoluteMotionEvent(InputInfoPtr pInfo, struct input_event *ev)
         pEvdev->abs_queued = 1;
     } else if (!pEvdev->mt_mask) {
         map = pEvdev->abs_axis_map[ev->code];
-        valuator_mask_set(pEvdev->vals, map, value);
-        pEvdev->abs_queued = 1;
+
+        /* check if the event must be translated into relative */
+        if (map < 2 && (pEvdev->flags & EVDEV_RELATIVE_MODE)) {
+            int oldval;
+            if (valuator_mask_fetch(pEvdev->old_vals, map, &oldval)) {
+                valuator_mask_set(pEvdev->rel_vals, map, value - oldval);
+                pEvdev->rel_queued = 1;
+            }
+            valuator_mask_set(pEvdev->old_vals, map, value);
+        } else {
+            /* the normal case: just store the number. */
+            valuator_mask_set(pEvdev->abs_vals, map, value);
+            pEvdev->abs_queued = 1;
+        }
     }
 }
 
@@ -850,6 +822,15 @@ EvdevProcessKeyEvent(InputInfoPtr pInfo, struct input_event *ev)
              * BTN_TOUCH as the proximity notifier */
             if (!pEvdev->use_proximity)
                 pEvdev->in_proximity = value ? ev->code : 0;
+            /* When !pEvdev->use_proximity, we don't report
+             * proximity events to the X server. However, we
+             * still want to keep track if one is in proximity or
+             * not. This is especially important for touchpad
+             * who report proximity information to the computer
+             * (but it is not sent to X) and who might send unreliable
+             * position information when not in_proximity.
+             */
+
             if (!(pEvdev->flags & (EVDEV_TOUCHSCREEN | EVDEV_TABLET)) ||
                 pEvdev->mt_mask)
                 break;
@@ -872,8 +853,8 @@ EvdevPostRelativeMotionEvents(InputInfoPtr pInfo)
 {
     EvdevPtr pEvdev = pInfo->private;
 
-    if (pEvdev->rel_queued) {
-        xf86PostMotionEventM(pInfo->dev, Relative, pEvdev->vals);
+    if (pEvdev->rel_queued && pEvdev->in_proximity) {
+        xf86PostMotionEventM(pInfo->dev, Relative, pEvdev->rel_vals);
     }
 }
 
@@ -895,7 +876,7 @@ EvdevPostAbsoluteMotionEvents(InputInfoPtr pInfo)
      * this scheme still just work.
      */
     if (pEvdev->abs_queued && pEvdev->in_proximity) {
-        xf86PostMotionEventM(pInfo->dev, Absolute, pEvdev->vals);
+        xf86PostMotionEventM(pInfo->dev, Absolute, pEvdev->abs_vals);
     }
 }
 
@@ -909,9 +890,7 @@ EvdevPostProximityEvents(InputInfoPtr pInfo, int which)
         switch (pEvdev->queue[i].type) {
             case EV_QUEUE_KEY:
             case EV_QUEUE_BTN:
-#ifdef MULTITOUCH
             case EV_QUEUE_TOUCH:
-#endif
                 break;
             case EV_QUEUE_PROXIMITY:
                 if (pEvdev->queue[i].val == which)
@@ -951,13 +930,11 @@ static void EvdevPostQueuedEvents(InputInfoPtr pInfo)
             break;
         case EV_QUEUE_PROXIMITY:
             break;
-#ifdef MULTITOUCH
         case EV_QUEUE_TOUCH:
             xf86PostTouchEvent(pInfo->dev, pEvdev->queue[i].detail.touch,
                                pEvdev->queue[i].val, 0,
                                pEvdev->queue[i].touchMask);
             break;
-#endif
         }
     }
 }
@@ -983,7 +960,6 @@ EvdevProcessSyncEvent(InputInfoPtr pInfo, struct input_event *ev)
     EvdevPostQueuedEvents(pInfo);
     EvdevPostProximityEvents(pInfo, FALSE);
 
-    memset(pEvdev->delta, 0, sizeof(pEvdev->delta));
     for (i = 0; i < ArrayLength(pEvdev->queue); i++)
     {
         EventQueuePtr queue = &pEvdev->queue[i];
@@ -993,8 +969,10 @@ EvdevProcessSyncEvent(InputInfoPtr pInfo, struct input_event *ev)
         /* don't reset the touchMask */
     }
 
-    if (pEvdev->vals)
-        valuator_mask_zero(pEvdev->vals);
+    if (pEvdev->rel_vals)
+        valuator_mask_zero(pEvdev->rel_vals);
+    if (pEvdev->abs_vals)
+        valuator_mask_zero(pEvdev->abs_vals);
     pEvdev->num_queue = 0;
     pEvdev->abs_queued = 0;
     pEvdev->rel_queued = 0;
@@ -1028,16 +1006,14 @@ EvdevProcessEvent(InputInfoPtr pInfo, struct input_event *ev)
 static void
 EvdevFreeMasks(EvdevPtr pEvdev)
 {
-#ifdef MULTITOUCH
     int i;
-#endif
 
     free(pEvdev->slots);
     pEvdev->slots = NULL;
-    valuator_mask_free(&pEvdev->vals);
+    valuator_mask_free(&pEvdev->abs_vals);
+    valuator_mask_free(&pEvdev->rel_vals);
     valuator_mask_free(&pEvdev->old_vals);
     valuator_mask_free(&pEvdev->prox);
-#ifdef MULTITOUCH
     valuator_mask_free(&pEvdev->mt_mask);
     if (pEvdev->last_mt_vals)
     {
@@ -1048,10 +1024,8 @@ EvdevFreeMasks(EvdevPtr pEvdev)
     }
     for (i = 0; i < EVDEV_MAXQUEUE; i++)
         valuator_mask_free(&pEvdev->queue[i].touchMask);
-#endif
 }
 
-#ifdef MULTITOUCH
 static void
 EvdevHandleMTDevEvent(InputInfoPtr pInfo, struct input_event *ev)
 {
@@ -1066,7 +1040,6 @@ EvdevHandleMTDevEvent(InputInfoPtr pInfo, struct input_event *ev)
         }
     }
 }
-#endif
 
 static void
 EvdevReadInput(InputInfoPtr pInfo)
@@ -1085,21 +1058,17 @@ EvdevReadInput(InputInfoPtr pInfo)
                                        strerror(-rc));
             break;
         } else if (rc == LIBEVDEV_READ_STATUS_SUCCESS) {
-#ifdef MULTITOUCH
             if (pEvdev->mtdev)
                 EvdevHandleMTDevEvent(pInfo, &ev);
             else
-#endif
                 EvdevProcessEvent(pInfo, &ev);
         }
         else { /* SYN_DROPPED */
             rc = libevdev_next_event(pEvdev->dev, LIBEVDEV_READ_FLAG_SYNC, &ev);
             while (rc == LIBEVDEV_READ_STATUS_SYNC) {
-#ifdef MULTITOUCH
                 if (pEvdev->mtdev)
                     EvdevHandleMTDevEvent(pInfo, &ev);
                 else
-#endif
                     EvdevProcessEvent(pInfo, &ev);
                 rc = libevdev_next_event(pEvdev->dev, LIBEVDEV_READ_FLAG_SYNC, &ev);
             }
@@ -1148,28 +1117,31 @@ static int
 EvdevAddKeyClass(DeviceIntPtr device)
 {
     int rc = Success;
-    XkbRMLVOSet rmlvo = {0};
+    XkbRMLVOSet rmlvo = {0},
+                defaults;
     InputInfoPtr pInfo;
 
     pInfo = device->public.devicePrivate;
 
+    XkbGetRulesDflts(&defaults);
+
     /* sorry, no rules change allowed for you */
     xf86ReplaceStrOption(pInfo->options, "xkb_rules", "evdev");
     rmlvo.rules = xf86SetStrOption(pInfo->options, "xkb_rules", NULL);
-    rmlvo.model = xf86SetStrOption(pInfo->options, "xkb_model", NULL);
-    rmlvo.layout = xf86SetStrOption(pInfo->options, "xkb_layout", NULL);
-    rmlvo.variant = xf86SetStrOption(pInfo->options, "xkb_variant", NULL);
-    rmlvo.options = xf86SetStrOption(pInfo->options, "xkb_options", NULL);
+    rmlvo.model = xf86SetStrOption(pInfo->options, "xkb_model", defaults.model);
+    rmlvo.layout = xf86SetStrOption(pInfo->options, "xkb_layout", defaults.layout);
+    rmlvo.variant = xf86SetStrOption(pInfo->options, "xkb_variant", defaults.variant);
+    rmlvo.options = xf86SetStrOption(pInfo->options, "xkb_options", defaults.options);
 
     if (!InitKeyboardDeviceStruct(device, &rmlvo, NULL, EvdevKbdCtrl))
         rc = !Success;
 
     XkbFreeRMLVOSet(&rmlvo, FALSE);
+    XkbFreeRMLVOSet(&defaults, FALSE);
 
     return rc;
 }
 
-#ifdef MULTITOUCH
 /* MT axes are counted twice - once as ABS_X (which the kernel keeps for
  * backwards compatibility), once as ABS_MT_POSITION_X. So we need to keep a
  * mapping of those axes to make sure we only count them once
@@ -1187,7 +1159,6 @@ static struct mt_axis_mappings mt_axis_mappings[] = {
     {ABS_MT_PRESSURE, ABS_PRESSURE},
     {ABS_MT_DISTANCE, ABS_DISTANCE},
 };
-#endif
 
 /**
  * return TRUE if the axis is not one we should count as true axis
@@ -1205,16 +1176,93 @@ is_blacklisted_axis(int axis)
     }
 }
 
+static int
+EvdevAddFakeSingleTouchAxes(InputInfoPtr pInfo)
+{
+    EvdevPtr pEvdev = pInfo->private;
+    int num_axes = 0;
+    int i;
+
+    if (pEvdev->fake_mt)
+        return 0;
+
+    /* Android drivers often have ABS_MT_POSITION_X but not ABS_X.
+       Loop over the MT->legacy axis table and add fake axes. */
+    for (i = 0; i < ArrayLength(mt_axis_mappings); i++)
+    {
+        int mt_code = mt_axis_mappings[i].mt_code;
+        int code = mt_axis_mappings[i].code;
+        const struct input_absinfo* abs;
+
+        if (!libevdev_has_event_code(pEvdev->dev, EV_ABS, mt_code) ||
+            libevdev_has_event_code(pEvdev->dev, EV_ABS, code))
+            continue;
+
+        abs = libevdev_get_abs_info(pEvdev->dev, mt_code);
+        if (libevdev_enable_event_code(pEvdev->dev, EV_ABS, code, abs))
+        {
+            xf86IDrvMsg(pInfo, X_ERROR, "Failed to fake axis %s.\n",
+                        libevdev_event_code_get_name(EV_ABS, code));
+            return -1;
+        }
+        xf86IDrvMsg(pInfo, X_INFO, "Faking axis %s.\n",
+                    libevdev_event_code_get_name(EV_ABS, code));
+        num_axes++;
+    }
+
+    return num_axes;
+}
+
+static void
+EvdevCountMTAxes(EvdevPtr pEvdev, int *num_mt_axes_total,
+                 int *num_mt_axes, int *num_axes)
+{
+    int axis;
+
+    if (pEvdev->fake_mt)
+        return;
+
+    /* Absolute multitouch axes: adjust mapping and axes counts. */
+    for (axis = ABS_MT_SLOT; axis < ABS_MAX; axis++)
+    {
+        int j;
+        Bool skip = FALSE;
+
+        if (!libevdev_has_event_code(pEvdev->dev, EV_ABS, axis))
+            continue;
+
+        /* Setup mapping if axis is in MT->legacy axis table. */
+        for (j = 0; j < ArrayLength(mt_axis_mappings); j++)
+        {
+            if (mt_axis_mappings[j].mt_code == axis &&
+                libevdev_has_event_code(pEvdev->dev, EV_ABS, mt_axis_mappings[j].code))
+            {
+                mt_axis_mappings[j].needs_mapping = TRUE;
+                skip = TRUE;
+            }
+        }
+
+        if (!is_blacklisted_axis(axis))
+        {
+            (*num_mt_axes_total)++;
+            if (!skip)
+                (*num_mt_axes)++;
+        }
+        (*num_axes)--;
+    }
+}
 
 static int
 EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
 {
     InputInfoPtr pInfo;
     EvdevPtr pEvdev;
-    int num_axes = 0, axis, i = 0;
+    int axis, i = 0;
+    int num_axes = 0; /* number of non-MT axes */
     int num_mt_axes = 0, /* number of MT-only axes */
         num_mt_axes_total = 0; /* total number of MT axes, including
                                   double-counted ones, excluding blacklisted */
+    int num_faked_axes;
     Atom *atoms;
     int mapping = 0;
 
@@ -1224,6 +1272,7 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
     if (!libevdev_has_event_type(pEvdev->dev, EV_ABS))
         goto out;
 
+    /* Find number of absolute axis, including MT ones, will decrease later. */
     for (i = 0; i < ABS_MAX; i++)
         if (libevdev_has_event_code(pEvdev->dev, EV_ABS, i))
             num_axes++;
@@ -1231,48 +1280,24 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
     if (num_axes < 1)
         goto out;
 
-#ifdef MULTITOUCH
-    for (axis = ABS_MT_SLOT; axis < ABS_MAX; axis++)
-    {
-        if (libevdev_has_event_code(pEvdev->dev, EV_ABS, axis))
-        {
-            int j;
-            Bool skip = FALSE;
+    num_faked_axes = EvdevAddFakeSingleTouchAxes(pInfo);
+    if (num_faked_axes < 0)
+        goto out;
 
-            for (j = 0; j < ArrayLength(mt_axis_mappings); j++)
-            {
-                if (mt_axis_mappings[j].mt_code == axis &&
-                    libevdev_has_event_code(pEvdev->dev, EV_ABS, mt_axis_mappings[j].code))
-                {
-                    mt_axis_mappings[j].needs_mapping = TRUE;
-                    skip = TRUE;
-                }
-            }
+    num_axes += num_faked_axes;
 
-            if (!is_blacklisted_axis(axis))
-            {
-                num_mt_axes_total++;
-                if (!skip)
-                    num_mt_axes++;
-            }
-            num_axes--;
-        }
-    }
+    EvdevCountMTAxes(pEvdev, &num_mt_axes_total, &num_mt_axes, &num_axes);
 
-    /* device only has mt-axes. the kernel should give us ABS_X etc for
-       backwards compat but some devices don't have it. */
+    /* Panic if, after faking ABS_X etc, we still only have mt-axes. */
     if (num_axes == 0 && num_mt_axes > 0) {
         xf86IDrvMsg(pInfo, X_ERROR,
                     "found only multitouch-axes. That shouldn't happen.\n");
         goto out;
     }
 
-#endif
 
 
-#ifdef HAVE_SMOOTH_SCROLLING
     num_axes += num_scroll_axes;
-#endif
 
     if (num_axes + num_mt_axes > MAX_VALUATORS) {
         xf86IDrvMsg(pInfo, X_WARNING, "found %d axes, limiting to %d.\n", num_axes, MAX_VALUATORS);
@@ -1287,14 +1312,19 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
 
     pEvdev->num_vals = num_axes;
     if (num_axes > 0) {
-        pEvdev->vals = valuator_mask_new(num_axes);
+        pEvdev->abs_vals = valuator_mask_new(num_axes);
         pEvdev->old_vals = valuator_mask_new(num_axes);
-        if (!pEvdev->vals || !pEvdev->old_vals) {
+        pEvdev->rel_vals = valuator_mask_new(num_axes);
+        /* One needs rel_vals for an absolute device because
+         *   a) their might be some (relative) scroll axes
+         *   b) the device could be set in EVDEV_RELATIVE_MODE
+         */
+        if (!pEvdev->abs_vals || !pEvdev->rel_vals || !pEvdev->old_vals) {
             xf86IDrvMsg(pInfo, X_ERROR, "failed to allocate valuator masks.\n");
             goto out;
         }
     }
-#ifdef MULTITOUCH
+
     if (num_mt_axes_total > 0) {
         int nslots = num_slots(pEvdev);
 
@@ -1345,14 +1375,11 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
             }
         }
     }
-#endif
     atoms = malloc((pEvdev->num_vals + num_mt_axes) * sizeof(Atom));
 
     i = 0;
     for (axis = ABS_X; i < MAX_VALUATORS && axis <= ABS_MAX; axis++) {
-#ifdef MULTITOUCH
         int j;
-#endif
         pEvdev->abs_axis_map[axis] = -1;
         if (!libevdev_has_event_code(pEvdev->dev, EV_ABS, axis) ||
             is_blacklisted_axis(axis))
@@ -1360,7 +1387,6 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
 
         mapping = i;
 
-#ifdef MULTITOUCH
         for (j = 0; j < ArrayLength(mt_axis_mappings); j++)
         {
             if (mt_axis_mappings[j].code == axis)
@@ -1369,13 +1395,11 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
                     mt_axis_mappings[j].needs_mapping)
                 mapping = mt_axis_mappings[j].mapping;
         }
-#endif
         pEvdev->abs_axis_map[axis] = mapping;
         if (mapping == i)
             i++;
     }
 
-#ifdef HAVE_SMOOTH_SCROLLING
     if (num_scroll_axes > 0)
     {
         mapping++; /* continue from abs axis mapping */
@@ -1387,7 +1411,6 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
         if (libevdev_has_event_code(pEvdev->dev, EV_REL, REL_WHEEL))
             pEvdev->rel_axis_map[REL_WHEEL] = mapping++;
     }
-#endif
 
     EvdevInitAxesLabels(pEvdev, Absolute, pEvdev->num_vals + num_mt_axes, atoms);
 
@@ -1397,7 +1420,6 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
         goto out;
     }
 
-#ifdef MULTITOUCH
     if (num_mt_axes_total > 0)
     {
         int num_touches = 0;
@@ -1425,7 +1447,6 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
             }
         }
     }
-#endif
 
     for (axis = ABS_X; axis < ABS_MT_SLOT; axis++) {
         const struct input_absinfo *abs;
@@ -1449,7 +1470,6 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
         xf86InitValuatorDefaults(device, axnum);
     }
 
-#ifdef MULTITOUCH
     for (axis = ABS_MT_TOUCH_MAJOR; axis <= ABS_MAX; axis++) {
         const struct input_absinfo *abs;
         int axnum = pEvdev->abs_axis_map[axis];
@@ -1483,9 +1503,7 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
                                    resolution, 0, resolution,
                                    Absolute);
     }
-#endif
 
-#ifdef HAVE_SMOOTH_SCROLLING
     if (num_scroll_axes)
     {
         int idx;
@@ -1531,7 +1549,6 @@ EvdevAddAbsValuatorClass(DeviceIntPtr device, int num_scroll_axes)
                               SCROLL_FLAG_NONE);
         }
     }
-#endif
 
     free(atoms);
 
@@ -1587,7 +1604,6 @@ out:
 static int
 EvdevSetScrollValuators(DeviceIntPtr device)
 {
-#ifdef HAVE_SMOOTH_SCROLLING
     InputInfoPtr pInfo;
     EvdevPtr pEvdev;
     int axnum;
@@ -1615,7 +1631,6 @@ EvdevSetScrollValuators(DeviceIntPtr device)
                           pEvdev->smoothScroll.horiz_delta,
                           SCROLL_FLAG_NONE);
     }
-#endif
 
     return Success;
 }
@@ -1648,9 +1663,7 @@ EvdevAddRelValuatorClass(DeviceIntPtr device, int num_scroll_axes)
         (num_scroll_axes == 0 || pEvdev->flags & EVDEV_ABSOLUTE_EVENTS))
             goto out;
 
-#ifdef HAVE_SMOOTH_SCROLLING
     num_axes += num_scroll_axes;
-#endif
 
     if (num_axes > MAX_VALUATORS) {
         xf86IDrvMsg(pInfo, X_WARNING, "found %d axes, limiting to %d.\n", num_axes, MAX_VALUATORS);
@@ -1659,8 +1672,8 @@ EvdevAddRelValuatorClass(DeviceIntPtr device, int num_scroll_axes)
 
     pEvdev->num_vals = num_axes;
     if (num_axes > 0) {
-        pEvdev->vals = valuator_mask_new(num_axes);
-        if (!pEvdev->vals)
+        pEvdev->rel_vals = valuator_mask_new(num_axes);
+        if (!pEvdev->rel_vals)
             goto out;
     }
     atoms = malloc(pEvdev->num_vals * sizeof(Atom));
@@ -1668,11 +1681,6 @@ EvdevAddRelValuatorClass(DeviceIntPtr device, int num_scroll_axes)
     for (axis = REL_X, map = 0; map < MAX_VALUATORS && axis <= REL_MAX; axis++)
     {
         pEvdev->rel_axis_map[axis] = -1;
-#ifndef HAVE_SMOOTH_SCROLLING
-        /* We don't post wheel events, so ignore them here too */
-        if (axis == REL_WHEEL || axis == REL_HWHEEL || axis == REL_DIAL)
-            continue;
-#endif
         if (!libevdev_has_event_code(pEvdev->dev, EV_REL, axis))
             continue;
         pEvdev->rel_axis_map[axis] = map;
@@ -1711,7 +1719,7 @@ EvdevAddRelValuatorClass(DeviceIntPtr device, int num_scroll_axes)
     return Success;
 
 out:
-    valuator_mask_free(&pEvdev->vals);
+    valuator_mask_free(&pEvdev->rel_vals);
     return !Success;
 }
 
@@ -1793,14 +1801,12 @@ EvdevCountScrollAxes(EvdevPtr pEvdev)
 {
     int num_scroll_axes = 0;
 
-#ifdef HAVE_SMOOTH_SCROLLING
     if (libevdev_has_event_code(pEvdev->dev, EV_REL, REL_WHEEL))
         num_scroll_axes++;
     if (libevdev_has_event_code(pEvdev->dev, EV_REL, REL_HWHEEL))
         num_scroll_axes++;
     if (libevdev_has_event_code(pEvdev->dev, EV_REL, REL_DIAL))
         num_scroll_axes++;
-#endif
 
     return num_scroll_axes;
 }
@@ -2201,14 +2207,16 @@ EvdevProbe(InputInfoPtr pInfo)
         }
     }
 
-#ifdef MULTITOUCH
     for (i = ABS_MT_SLOT; i < ABS_MAX; i++) {
         if (libevdev_has_event_code(pEvdev->dev, EV_ABS, i)) {
             has_mt = TRUE;
             break;
         }
     }
-#endif
+
+    if (libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_MT_SLOT) &&
+        libevdev_get_num_slots(pEvdev->dev) == -1)
+        pEvdev->fake_mt = TRUE;
 
     if (ignore_abs && has_abs_axes)
     {
@@ -2231,6 +2239,8 @@ EvdevProbe(InputInfoPtr pInfo)
                     pEvdev->flags |= EVDEV_BUTTON_EVENTS;
                 }
             }
+            if (pEvdev->fake_mt)
+                xf86IDrvMsg(pInfo, X_PROBED, "Fake MT device detected\n");
         }
 
         if ((libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_X) &&
@@ -2265,10 +2275,8 @@ EvdevProbe(InputInfoPtr pInfo)
                     pEvdev->flags |= EVDEV_BUTTON_EVENTS;
             }
         } else {
-#ifdef MULTITOUCH
             if (!libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_MT_POSITION_X) ||
                 !libevdev_has_event_code(pEvdev->dev, EV_ABS, ABS_MT_POSITION_Y))
-#endif
                 EvdevForceXY(pInfo, Absolute);
         }
     }
@@ -2291,6 +2299,12 @@ EvdevProbe(InputInfoPtr pInfo)
         pEvdev->invert_y = xf86SetBoolOption(pInfo->options, "InvertY", FALSE);
         pEvdev->swap_axes = xf86SetBoolOption(pInfo->options, "SwapAxes", FALSE);
 
+        pEvdev->resolution = xf86SetIntOption(pInfo->options, "Resolution", 0);
+        if (pEvdev->resolution < 0) {
+            xf86IDrvMsg(pInfo, X_ERROR, "Resolution must be a positive number");
+            pEvdev->resolution = 0;
+        }
+
         str = xf86CheckStrOption(pInfo->options, "Calibration", NULL);
         if (str) {
             num_calibration = sscanf(str, "%d %d %d %d",
@@ -2342,14 +2356,12 @@ EvdevProbe(InputInfoPtr pInfo)
         pEvdev->flags |= EVDEV_BUTTON_EVENTS;
         pEvdev->flags |= EVDEV_RELATIVE_EVENTS;
 
-#ifdef HAVE_SMOOTH_SCROLLING
         pEvdev->smoothScroll.vert_delta =
             xf86SetIntOption(pInfo->options, "VertScrollDelta", 1);
         pEvdev->smoothScroll.horiz_delta =
             xf86SetIntOption(pInfo->options, "HorizScrollDelta", 1);
         pEvdev->smoothScroll.dial_delta =
             xf86SetIntOption(pInfo->options, "DialDelta", 1);
-#endif
     }
 
 out:
@@ -2379,7 +2391,6 @@ EvdevSetCalibration(InputInfoPtr pInfo, int num_calibration, int calibration[4])
     }
 }
 
-#ifdef MULTITOUCH
 /**
  * Open an mtdev device for this device. mtdev is a bit too generous with
  * memory usage, so only do so for multitouch protocol A devices.
@@ -2428,7 +2439,6 @@ EvdevOpenMTDev(InputInfoPtr pInfo)
 
     return TRUE;
 }
-#endif
 
 static int
 EvdevOpenDevice(InputInfoPtr pInfo)
@@ -2486,13 +2496,11 @@ EvdevOpenDevice(InputInfoPtr pInfo)
         return BadMatch;
     }
 
-#ifdef MULTITOUCH
     if (!EvdevOpenMTDev(pInfo)) {
         xf86Msg(X_ERROR, "%s: Couldn't open mtdev device\n", pInfo->name);
         EvdevCloseDevice(pInfo);
         return BadValue;
     }
-#endif
 
     return Success;
 }
@@ -2507,13 +2515,11 @@ EvdevCloseDevice(InputInfoPtr pInfo)
         pInfo->fd = -1;
     }
 
-#ifdef MULTITOUCH
     if (pEvdev->mtdev)
     {
         mtdev_close_delete(pEvdev->mtdev);
         pEvdev->mtdev = NULL;
     }
-#endif
 
 }
 
@@ -2558,9 +2564,7 @@ EvdevAlloc(InputInfoPtr pInfo)
     pEvdev->in_proximity = 1;
     pEvdev->use_proximity = 1;
 
-#ifdef MULTITOUCH
     pEvdev->cur_slot = -1;
-#endif
 
     for (i = 0; i < ArrayLength(pEvdev->rel_axis_map); i++)
         pEvdev->rel_axis_map[i] = -1;
@@ -2647,7 +2651,7 @@ _X_EXPORT InputDriverRec EVDEV = {
     EvdevPreInit,
     EvdevUnInit,
     NULL,
-    evdevDefaults,
+    NULL,
 #ifdef XI86_DRV_CAP_SERVER_FD
     XI86_DRV_CAP_SERVER_FD
 #endif
@@ -2767,22 +2771,22 @@ static void EvdevInitButtonLabels(EvdevPtr pEvdev, int natoms, Atom *atoms)
 
     for (button = BTN_MISC; button < BTN_JOYSTICK; button++)
     {
-        if (libevdev_has_event_code(pEvdev->dev, EV_KEY, button))
-        {
-            int group = (button % 0x100)/16;
-            int idx = button - ((button/16) * 16);
+        int group = (button % 0x100)/16;
+        int idx = button - ((button/16) * 16);
 
-            if (!btn_labels[group][idx])
-                continue;
+        if (!libevdev_has_event_code(pEvdev->dev, EV_KEY, button))
+            continue;
 
-            atom = XIGetKnownProperty(btn_labels[group][idx]);
-            if (!atom)
-                continue;
+        if (!btn_labels[group][idx])
+            continue;
 
-            /* Props are 0-indexed, button numbers start with 1 */
-            bmap = EvdevUtilButtonEventToButtonNumber(pEvdev, button) - 1;
-            atoms[bmap] = atom;
-        }
+        atom = XIGetKnownProperty(btn_labels[group][idx]);
+        if (!atom)
+            continue;
+
+        /* Props are 0-indexed, button numbers start with 1 */
+        bmap = EvdevUtilButtonEventToButtonNumber(pEvdev, button) - 1;
+        atoms[bmap] = atom;
     }
 
     /* wheel buttons, hardcoded anyway */
@@ -2836,6 +2840,9 @@ EvdevInitProperty(DeviceIntPtr dev)
                                 strlen(XI_PROP_VIRTUAL_DEVICE), TRUE);
         rc = XIChangeDeviceProperty(dev, prop_virtual, XA_INTEGER, 8,
                                     PropModeReplace, 1, &virtual, FALSE);
+        if (rc != Success)
+            return;
+
         XISetDevicePropertyDeletable(dev, prop_virtual, FALSE);
     }
 
@@ -2908,8 +2915,11 @@ EvdevInitProperty(DeviceIntPtr dev)
             }
 
             EvdevInitAxesLabels(pEvdev, mode, num_axes, atoms);
-            XIChangeDeviceProperty(dev, prop_axis_label, XA_ATOM, 32,
-                                   PropModeReplace, num_axes, atoms, FALSE);
+            rc = XIChangeDeviceProperty(dev, prop_axis_label, XA_ATOM, 32,
+                                        PropModeReplace, num_axes, atoms, FALSE);
+            if (rc != Success)
+                return;
+
             XISetDevicePropertyDeletable(dev, prop_axis_label, FALSE);
         }
         /* Button labelling */
@@ -2917,12 +2927,14 @@ EvdevInitProperty(DeviceIntPtr dev)
         {
             Atom atoms[EVDEV_MAXBUTTONS];
             EvdevInitButtonLabels(pEvdev, EVDEV_MAXBUTTONS, atoms);
-            XIChangeDeviceProperty(dev, prop_btn_label, XA_ATOM, 32,
-                                   PropModeReplace, pEvdev->num_buttons, atoms, FALSE);
+            rc = XIChangeDeviceProperty(dev, prop_btn_label, XA_ATOM, 32,
+                                        PropModeReplace, pEvdev->num_buttons, atoms, FALSE);
+            if (rc != Success)
+                return;
+
             XISetDevicePropertyDeletable(dev, prop_btn_label, FALSE);
         }
 
-#ifdef HAVE_SMOOTH_SCROLLING
         {
             int smooth_scroll_values[3] = {
                 pEvdev->smoothScroll.vert_delta,
@@ -2931,11 +2943,13 @@ EvdevInitProperty(DeviceIntPtr dev)
             };
             prop_scroll_dist = MakeAtom(EVDEV_PROP_SCROLL_DISTANCE,
                                         strlen(EVDEV_PROP_SCROLL_DISTANCE), TRUE);
-            XIChangeDeviceProperty(dev, prop_scroll_dist, XA_INTEGER, 32,
-                                   PropModeReplace, 3, smooth_scroll_values, FALSE);
+            rc = XIChangeDeviceProperty(dev, prop_scroll_dist, XA_INTEGER, 32,
+                                        PropModeReplace, 3, smooth_scroll_values, FALSE);
+            if (rc != Success)
+                return;
+
             XISetDevicePropertyDeletable(dev, prop_scroll_dist, FALSE);
         }
-#endif
 
     }
 
